package ij.process;
import java.awt.*;

/** This class does bit blitting of 16-bit images. */
public class ShortBlitter implements Blitter {

	static final int MIN_SHORT = -32768;
	static final int MAX_SHORT = 32767;
	static final int UNSIGNED_ADD = ADD+20;
	static final int UNSIGNED_SUBTRACT = SUBTRACT+20;
	static final int UNSIGNED_MULTIPLY = MULTIPLY+20;
	static final int UNSIGNED_DIVIDE = DIVIDE+20;
	static final int UNSIGNED_DIFFERENCE = DIFFERENCE+20;
	
	private ShortProcessor ip;
	private int width, height;
	private short[] pixels;
	
	public void setTransparentColor(Color c) {
	}

	/** Constructs a ShortBlitter from a ShortProcessor. */
	public ShortBlitter(ShortProcessor ip) {
		this.ip = ip;
		width = ip.getWidth();
		height = ip.getHeight();
		pixels = (short[])ip.getPixels();
	}

	/** Copies the 16-bit image in 'ip' to (x,y) using the specified mode. */
	public void copyBits(ImageProcessor ip, int xloc, int yloc, int mode) {
		Rectangle r1, r2;
		int srcIndex, dstIndex;
		int xSrcBase, ySrcBase;
		short[] srcPixels;
		
		int srcWidth = ip.getWidth();
		int srcHeight = ip.getHeight();
		r1 = new Rectangle(srcWidth, srcHeight);
		r1.setLocation(xloc, yloc);
		r2 = new Rectangle(width, height);
		if (!r1.intersects(r2))
			return;
		srcPixels = (short [])ip.getPixels();
		r1 = r1.intersection(r2);
		xSrcBase = (xloc<0)?-xloc:0;
		ySrcBase = (yloc<0)?-yloc:0;
		int src, dst;
		boolean unsigned = ip.isUnsigned();
		if (unsigned)
			switch (mode) {
				case ADD: mode = UNSIGNED_ADD; break;
				case SUBTRACT: mode = UNSIGNED_SUBTRACT; break;
				case MULTIPLY: mode = UNSIGNED_MULTIPLY; break;
				case DIVIDE: mode = UNSIGNED_DIVIDE; break;
				case DIFFERENCE: mode = UNSIGNED_DIFFERENCE; break;
				default:
			}
		for (int y=r1.y; y<(r1.y+r1.height); y++) {
			srcIndex = (y-yloc)*srcWidth + (r1.x-xloc);
			dstIndex = y * width + r1.x;
			switch (mode) {
				case COPY: case COPY_INVERTED: case COPY_TRANSPARENT:
					for (int i=r1.width; --i>=0;)
						pixels[dstIndex++] = srcPixels[srcIndex++];
					break;
				case ADD:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]+pixels[dstIndex];
						if (dst<MIN_SHORT) dst = MIN_SHORT;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case UNSIGNED_ADD:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]+pixels[dstIndex]+32768;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case AVERAGE:
					for (int i=r1.width; --i>=0;) {
						dst = (srcPixels[srcIndex++]+pixels[dstIndex])/2;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case DIFFERENCE:
					for (int i=r1.width; --i>=0;) {
						dst = pixels[dstIndex]-srcPixels[srcIndex++];
						if (dst<0) dst = -dst;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case UNSIGNED_DIFFERENCE:
					for (int i=r1.width; --i>=0;) {
						dst = pixels[dstIndex]-srcPixels[srcIndex++];
						if (dst<0) dst = -dst;
						pixels[dstIndex++] = (short)(dst-32768);
					}
					break;
				case SUBTRACT:
					for (int i=r1.width; --i>=0;) {
						dst = pixels[dstIndex]-srcPixels[srcIndex++];
						if (dst<MIN_SHORT) dst = MIN_SHORT;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case UNSIGNED_SUBTRACT:
					for (int i=r1.width; --i>=0;) {
						dst = pixels[dstIndex]-srcPixels[srcIndex++];
						dst -= 32768;
						if (dst<MIN_SHORT) dst = MIN_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case MULTIPLY:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]*pixels[dstIndex];
						if (dst<MIN_SHORT) dst = MIN_SHORT;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case UNSIGNED_MULTIPLY:
					for (int i=r1.width; --i>=0;) {
						src = srcPixels[srcIndex++]+32768;
						dst = pixels[dstIndex]+32768;
						dst = src*dst-32768;
						if (dst>MAX_SHORT) dst = MAX_SHORT;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case DIVIDE:
					for (int i=r1.width; --i>=0;) {
						src = srcPixels[srcIndex++];
						if (src==0)
							dst = MAX_SHORT;
						else
							dst = pixels[dstIndex]/src;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case UNSIGNED_DIVIDE:
					for (int i=r1.width; --i>=0;) {
						src = srcPixels[srcIndex++]+32768;
						if (src==0)
							dst = MAX_SHORT;
						else
							dst = (pixels[dstIndex]+32768)/src;
						pixels[dstIndex++] = (short)(dst-32768);
					}
					break;
				case AND:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]&pixels[dstIndex];
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case OR:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]|pixels[dstIndex];
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case XOR:
					for (int i=r1.width; --i>=0;) {
						dst = srcPixels[srcIndex++]^pixels[dstIndex];
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case MIN:
					for (int i=r1.width; --i>=0;) {
						src = srcPixels[srcIndex++];
						dst = pixels[dstIndex];
						if (src<dst) dst = src;
						pixels[dstIndex++] = (short)dst;
					}
					break;
				case MAX:
					for (int i=r1.width; --i>=0;) {
						src = srcPixels[srcIndex++];
						dst = pixels[dstIndex];
						if (src>dst) dst = src;
						pixels[dstIndex++] = (short)dst;
					}
					break;
			}
			if (y%20==0)
				ip.showProgress((double)(y-r1.y)/r1.height);
		}
		ip.hideProgress();
	}
}